package meerkat.bulletinboard.sqlserver;

import java.sql.*;
import java.util.*;

import com.google.protobuf.InvalidProtocolBufferException;
import com.google.protobuf.ProtocolStringList;

import meerkat.bulletinboard.BulletinBoardServer;
import meerkat.comm.CommunicationException;
import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Crypto.Signature;
import meerkat.protobuf.Crypto.SignatureVerificationKey;
import meerkat.crypto.Digest;
import meerkat.crypto.concrete.SHA256Digest;

/**
 * This is a generic SQL implementation of the BulletinBoardServer API.
 */
public class BulletinBoardSQLServer implements BulletinBoardServer{

	/**
	 * This interface provides the required implementation-specific data to enable an access to an actual SQL server.
	 * It accounts for differences in languages between SQL DB types and for the different addresses needed to access them.
	 */
	public interface SQLQueryProvider {

		/**
		 * Allowed query types.
		 * Note that each query returned has to comply with the placeholder ("?")  requirements written in its comment.
		 */
		public static enum QueryType {
			FIND_MSG_ID,		// Placeholders for: MsgId
			INSERT_MSG,			// Placeholders for: MsgId, Msg
			INSERT_NEW_TAG,		// Placeholders for: Tag
			CONNECT_TAG,		// Placeholders for: EntryNum, Tag
			ADD_SIGNATURE,		// Placeholders for: EntryNum, SignerId, Signature
			GET_SIGNATURES,		// Placeholders for: EntryNum
			GET_MESSAGES		// Placeholders for: N/A
		}

		/**
		 * This function translates a QueryType into an actual SQL query.
		 * @param queryType is the type of query requested
         * @return a string representation of the query for the specific type of SQL database implemented.
         */
		public String getSQLString(QueryType queryType) throws IllegalArgumentException;

		public String getCondition(FilterType filterType) throws IllegalArgumentException;

		/**
		 * @return the string needed in order to connect to the DB.
         */
		public String getConnectionString();

		/**
		 * This is used to get a list of queries that together create the schema needed for the DB.
		 * Note that these queries should not assume anything about the current state of the DB.
		 * In particular: they should not erase any existing tables and/or entries.
		 * @return The list of queries.
         */
		public List<String> getSchemaCreationCommands();

		/**
		 * This is used to get a list of queries that together delete the schema needed for the DB.
		 * This is useful primarily for tests, in which we want to make sure we start with a clean DB.
		 * @return The list of queries.
		 */
		public List<String> getSchemaDeletionCommands();

	}

	/**
	 * This class implements a comparator for the MessageFilter class
	 * The comparison is done solely by comparing the type of the filter
	 * This is used to sort the filters by type
	 */
	public class FilterTypeComparator implements Comparator<MessageFilter> {

		@Override
		public int compare(MessageFilter filter1, MessageFilter filter2) {
			return filter1.getTypeValue() - filter2.getTypeValue();
		}
	}

	protected SQLQueryProvider sqlQueryProvider;

	protected Connection connection;
	
	protected Digest digest;
	
	protected List<SignatureVerificationKey> trusteeSignatureVerificationArray;
	protected int minTrusteeSignatures;

	protected List<List<SignatureVerificationKey>> pollingCommitteeSignatureVerificationKeyArray;
	protected int minCommiteeSignatures;

	/**
	 * This constructor sets the type of SQL language in use.
	 * @param sqlQueryProvider is the provider of the SQL query strings required for actual operation of the server.
     */
	public BulletinBoardSQLServer(SQLQueryProvider sqlQueryProvider) {
		this.sqlQueryProvider = sqlQueryProvider;
	}

	private void createSchema() throws SQLException {

		final int TIMEOUT = 20;

		Statement statement = connection.createStatement();
		statement.setQueryTimeout(TIMEOUT);

		for (String command : sqlQueryProvider.getSchemaCreationCommands()) {
			statement.executeUpdate(command);
		}

		statement.close();
	}

	/**
	 * This method initializes the signatures, connects to the DB and creates the schema (if required).
	 */
	@Override
	public void init(String meerkatDB) throws CommunicationException {
		// TODO write signature reading part.
		
		digest = new SHA256Digest();

		try{

			String dbString = sqlQueryProvider.getConnectionString();
			connection = DriverManager.getConnection(dbString);
		} catch (SQLException e) {
			throw new CommunicationException("Couldn't form a connection with the database " + e.getMessage());
		}

		try {
			createSchema();
		} catch (SQLException e) {
			throw new CommunicationException("Couldn't create schema " + e.getMessage());
		}

	}

	/**
	 * This method verifies the authenticity of the received message based on
	 * the stored signatures.
	 * 
	 * @param msg
	 *            is the message to authenticate (containing the signature).
	 * @return TRUE if the message is authenticated and FALSE otherwise.
	 */
	private boolean verifyMessage(BulletinBoardMessage msg) {
		//TODO: Replace with actual verification.
		return true;
	}
	
	/**
	 * This procedure makes sure that all tags in the given list have an entry in the tags table. 
	 * @param tags
	 */
	protected void insertNewTags(String[] tags) throws SQLException {

		PreparedStatement pstmt;
		String sql;

		try {

			sql = sqlQueryProvider.getSQLString(SQLQueryProvider.QueryType.INSERT_NEW_TAG);
			pstmt = connection.prepareStatement(sql);

			for (String tag : tags){
				pstmt.setString(1, tag);
				pstmt.addBatch();
			}

			pstmt.executeBatch();
			pstmt.close();

		} catch (SQLException e){
			throw new SQLException("Error adding new tags to table: " + e.getMessage());
		}

	}
	
	/**
	 * This procedure is used to convert a boolean to a BoolMsg.
	 * @param b is the boolean to convert.
	 * @return a ProtoBuf message with boolean payload.
	 */
	private BoolMsg boolToBoolMsg(boolean b){
		return BoolMsg.newBuilder()
			.setValue(b)
			.build();
	}

	@Override
	public BoolMsg postMessage(BulletinBoardMessage msg) throws CommunicationException {
		
		if (!verifyMessage(msg)) {
			return boolToBoolMsg(false);
		}
		
		PreparedStatement pstmt;
		ResultSet rs;
		String sql;
		byte[] msgID;
		long entryNum;
		
		ProtocolStringList tagList;
		String[] tags;
		
		List<Signature> signatureList;
		Signature[] signatures;
		
		// Calculate message ID (depending only on the the unsigned message)
		
		digest.reset();
		digest.update(msg.getMsg());
		
		msgID = digest.digest();
		
		// Add message to table if needed and store entry number of message.
		
		try {
			
			sql = sqlQueryProvider.getSQLString(SQLQueryProvider.QueryType.FIND_MSG_ID);
			pstmt = connection.prepareStatement(sql);
			pstmt.setBytes(1, msgID);
			rs = pstmt.executeQuery();
			
			if (rs.next()){
				
				entryNum = rs.getLong(1);
				
			} else{
				
				sql = sqlQueryProvider.getSQLString(SQLQueryProvider.QueryType.INSERT_MSG);
			    pstmt = connection.prepareStatement(sql, Statement.RETURN_GENERATED_KEYS);
			    pstmt.setBytes(1, msgID);
			    pstmt.setBytes(2, msg.getMsg().toByteArray());
			    pstmt.executeUpdate();
			    
			    rs = pstmt.getGeneratedKeys();
			    rs.next();
			    entryNum = rs.getLong(1);
				
			}
		    
		    pstmt.close();
		    
		} catch (SQLException e) {
			throw new CommunicationException("Error inserting into MsgTable: " + e.getMessage());
		}
		
		// Retrieve tags and store new ones in tag table.
		
		try {
			
		    tagList = msg.getMsg().getTagList();
		    tags = new String[tagList.size()];
		    tags = tagList.toArray(tags);
		    
		    insertNewTags(tags);
		    
		} catch (SQLException e) {
			throw new CommunicationException(e.getMessage());
		}
		
		// Connect message to tags.
		
		try{
		    sql = sqlQueryProvider.getSQLString(SQLQueryProvider.QueryType.CONNECT_TAG);
	    	pstmt = connection.prepareStatement(sql);
		    
	    	pstmt.setLong(1, entryNum);
	    	
		    for (String tag : tags){
		    	pstmt.setString(2, tag);
		    	pstmt.addBatch();
		    }
		    
		    pstmt.executeBatch();
		    pstmt.close();
			
		} catch (SQLException e) {
			throw new CommunicationException("Error Linking tags: " + e.getMessage());
		}
		
		// Retrieve signatures.
		
	    signatureList = msg.getSigList();
	    signatures = new Signature[signatureList.size()];
	    signatures = signatureList.toArray(signatures);
		
		// Connect message to signatures.
		
		try{
		    sql = sqlQueryProvider.getSQLString(SQLQueryProvider.QueryType.ADD_SIGNATURE);
	    	pstmt = connection.prepareStatement(sql);
		    
	    	pstmt.setLong(1, entryNum);
	    	
		    for (Signature sig : signatures){
		    	
		    	pstmt.setBytes(2, sig.getSignerId().toByteArray());
		    	pstmt.setBytes(3, sig.toByteArray());
		    	pstmt.addBatch();
		    }
		    
		    pstmt.executeBatch();
		    pstmt.close();
			
		} catch (SQLException e) {
			throw new CommunicationException("Error Linking tags: " + e.getMessage());
		}

		return boolToBoolMsg(true);
	}

	@Override
	public BulletinBoardMessageList readMessages(MessageFilterList filterList) throws CommunicationException {
		PreparedStatement pstmt;
		ResultSet messages, signatures;

		long entryNum;
		BulletinBoardMessageList.Builder resultListBuilder = BulletinBoardMessageList.newBuilder();
		BulletinBoardMessage.Builder messageBuilder;

		String sql;

		List<MessageFilter> filters = new ArrayList<MessageFilter>(filterList.getFilterList());
		int i;

		boolean tagsRequired = false;
		boolean signaturesRequired = false;

		boolean isFirstFilter = true;

		Collections.sort(filters, new FilterTypeComparator());

		// Check if Tag/Signature tables are required for filtering purposes.

		sql = sqlQueryProvider.getSQLString(SQLQueryProvider.QueryType.GET_MESSAGES);

		// Add conditions.

		if (!filters.isEmpty()){
			sql += " WHERE ";

			for (MessageFilter filter : filters){

				if (filter.getType().getNumber() != FilterType.MAX_MESSAGES_VALUE){
					if (isFirstFilter){
						isFirstFilter = false;
					} else{
						sql += " AND ";
					}
				}

				sql += sqlQueryProvider.getCondition(filter.getType());
			}

		}

		// Make query.

		try {
			pstmt = connection.prepareStatement(sql);

			// Specify values for filters.

			i = 1;
			for (MessageFilter filter : filters){

				switch (filter.getType().getNumber()){

					case FilterType.EXACT_ENTRY_VALUE: // Go through.
					case FilterType.MAX_ENTRY_VALUE:
						pstmt.setLong(i, filter.getEntry());
						i++;
						break;

					case FilterType.MSG_ID_VALUE: // Go through.
					case FilterType.SIGNER_ID_VALUE:
						pstmt.setBytes(i, filter.getId().toByteArray());
						i++;
						break;

					case FilterType.TAG_VALUE:
						pstmt.setString(i, filter.getTag());
						i++;
						break;

					// The max-messages condition is applied as a suffix. Therefore, it is treated differently.
					case FilterType.MAX_MESSAGES_VALUE:
						pstmt.setLong(filters.size(), filter.getMaxMessages());
						i++;
						break;

				}
			}

			// Run query.

			messages = pstmt.executeQuery();

			// Compile list of messages.

			sql = sqlQueryProvider.getSQLString(SQLQueryProvider.QueryType.GET_SIGNATURES);
			pstmt = connection.prepareStatement(sql);

			while (messages.next()){

				// Get entry number and retrieve signatures.

				entryNum = messages.getLong(1);
				pstmt.setLong(1, entryNum);
				signatures = pstmt.executeQuery();

				// Create message and append signatures.

				messageBuilder = BulletinBoardMessage.newBuilder()
						.setEntryNum(entryNum)
						.setMsg(UnsignedBulletinBoardMessage.parseFrom(messages.getBytes(2)));

				while (signatures.next()){
					messageBuilder.addSig(Signature.parseFrom(signatures.getBytes(1)));
				}

				// Finalize message and add to message list.

				resultListBuilder.addMessage(messageBuilder.build());

			}

			pstmt.close();

		} catch (SQLException e){
			throw new CommunicationException("Error reading messages from DB: " + e.getMessage());
		} catch (InvalidProtocolBufferException e) {
			throw new CommunicationException("Invalid data from DB: " + e.getMessage());
		}

		//Combine results and return.

		return resultListBuilder.build();
	}

	@Override
	public void close() throws CommunicationException {

		try{
			connection.close();
		} catch (SQLException e) {

			throw new CommunicationException("Couldn't close connection to the database");

		}

	}

}
